@@ -2,12 +2,13 @@
 """
 A test harness for gnupg.py.
 
-Copyright (C) 2008-2014 Vinay Sajip. All rights reserved.
+Copyright (C) 2008-2016 Vinay Sajip. All rights reserved.
 """
 import doctest
 import logging
 import os.path
 import os
+import re
 import shutil
 import stat
 import sys
@@ -17,7 +18,7 @@ import unittest
 import gnupg
 
 __author__ = "Vinay Sajip"
-__date__  = "$24-Sep-2015 18:05:15$"
+__date__  = "$10-Sep-2016 08:38:57$"
 
 ALL_TESTS = True
 
@@ -80,16 +81,71 @@ WdPQjKEfKnr+bW4yubwMUYKyAJ4uiE8Rv/oEED1oM3xeJqa+MJ9V1w==
 =sqld
 -----END PGP PUBLIC KEY BLOCK-----"""
 
+SIGNED_KEYS="""-----BEGIN PGP PUBLIC KEY BLOCK-----
+Version: GnuPG v2
+
+mI0EVcnKUQEEAKWazmfM0kbvDdw7Kos2NARaX67c8iJ3GOBimUvYLj4VR3Mqrm34
+ZdLlS8jCmid+qoisefvGW5uw5Q3gIs0mdEdUpFKlXNiIja/Dg/FHjjJPPCjfzDTh
+Q03EYA7QvOnXZXhYPBqK7NitsNXW4lPnIJdanLx7yMuL+2Xb+tF39mwnABEBAAG0
+LUpvc2h1YSBDYWx2ZXJ0IChBIHRlc3QgdXNlcikgPGpjQGV4YW1wbGUuY29tPoi3
+BBMBCAAhBQJVycpRAhsDBQsJCAcCBhUICQoLAgQWAgMBAh4BAheAAAoJELxvNQ+z
+0EB2jcED/0lHKaEkyd6cj0Zckf9luIkZ4Hno/vRCquTI7c3aPjS3qmE8mOvKSBCV
++SamPdRM7DdjkdBrrKy2HtiDqbM+1/CdXuQka2SlJWyLCJe48+KWfBpqlY3N4t53
+JjHRitDB+hC8njWTV5prli6EgsBPAF+ZkO0iZhlsMmWdDWgqDpGRiJwEEAEIAAYF
+AlXJym8ACgkQBXzPZYwHT9oiiQQAvPF8ubwRopnXIMDQgSxKyFDM1MI1w/wb4Okd
+/MkMeZSmdcHJ6pEymp5bYciCBuLW+jw0vZWza3YloO/HtuppnF6A9a1UvYcp/diI
+O5qkQqYPlui1PJl7hQ014ioniMfOcC4X/r6PDbC78Pczje0Yh9AOqNGeCyNyNdlc
+pjaHb0m4jQRVycpRAQQAo9JjW75F5wTVVO552cGCZWqZvDyBt9+IkoK9Bc+ggdn5
+6R8QVCihYuaSzcSEN84zHaR3MmGKHraCmCSlfe7w0d41Dlns0P03KMdIZOGrm045
+F8TXdSSPQOv5tA4bz3k2lGD0zB8l4NUWFaZ5fzw2i73FF4O/FwCU8xd/JCKVPkkA
+EQEAAYifBBgBCAAJBQJVycpRAhsMAAoJELxvNQ+z0EB2xLYD/i3tKirQlVB+32WP
+wggstqDp1BlUBmDb+4Gndpg4l7omJTTyOsF26SbYgXZqAdEd5T/UfpEla0DKiBYh
+2/CFYXadkgX/ME+GTetTmD4hHoBNmdXau92buXsIXkwh+JR+RC3cl2U6tWb/MIRd
+zvJiok8W8/FT/QrEjIa2etN2d+KR
+=nNBX
+-----END PGP PUBLIC KEY BLOCK-----
+-----BEGIN PGP PUBLIC KEY BLOCK-----
+Version: GnuPG v2
+
+mI0EVcnKNgEEANIVlIUyRXWHP/ljdMEA8B5NxecRCKusUIPxeapk2do5UCZgR1q8
+5wOP4K/+W3Uj85ylOOCNTFYKRozAHsPMAmQ38W93DZYqFbG6d7rwMvz4pVe0wUtj
+SBINoKnoEDZwx3erxFKOkp/5fF3NoYSIx9a0Ds21ESk0TAuH5Tg934YhABEBAAG0
+MVdpbnN0b24gU21pdGggKEEgdGVzdCB1c2VyKSA8d2luc3RvbkBleGFtcGxlLmNv
+bT6ItwQTAQgAIQUCVcnKNgIbAwULCQgHAgYVCAkKCwIEFgIDAQIeAQIXgAAKCRAF
+fM9ljAdP2h05A/4vmnxV1MwcOhJTHZys5g2/j5UoZG7V7lPGpJaojSAIVzYXZtwT
+5A7OY8Nl21kIY6gnZlgbTRpHN8Qq2wRKAyW5o6wQvuN16CW4bmGjoHYRGPqkeM0w
+G40W/v88JXrYDNNe/68g4pnPsZ3J0oMLbRvCaDQQHXBuZNJrT1sOxl9Of7iNBFXJ
+yjYBBACmHbs0PdOF8NEGc+fEtmdKOSKOkrcvg1wTu1KFFTBFEbseHOCNpx+R6lfO
+ZiZmHGdKeJhTherfjHaY5jmvyDWq5TLZXK61quNsWxmY2zJ0SRwrIG/CWi4bMi5t
+JNc23vMumkz4X5g7x0Ea7xEWkcYBn0H6sZDAtb8d8mrlWkMekQARAQABiJ8EGAEI
+AAkFAlXJyjYCGwwACgkQBXzPZYwHT9pQIwP8D9/VroykSE2J3gy0S6HC287jXqXF
+0zWejUAQtWUSSRx4esqfLE8lfae6+LDHO8D0Bf6YUJmu7ATOZP2/TIas7JrNvXWc
+NKWl2MHEAGUYq8utCjZ3dKKhaV7UvcY4PyLIpFteNkOz4wFe6C0Mm+1NYwokIFyh
+zPBq9eFk7Xx9Wrc=
+=HT6N
+-----END PGP PUBLIC KEY BLOCK-----
+"""
+
+
 def is_list_with_len(o, n):
     return isinstance(o, list) and len(o) == n
 
+BASE64_PATTERN = re.compile(r'^(?:[A-Z0-9+/]{4})*(?:[A-Z0-9+/]{2}==|[A-Z0-9+/]{3}=)?$', re.I)
+
+def get_key_data(s):
+    lines = s.split('\n')
+    result = ''
+    for line in lines:
+        m = BASE64_PATTERN.match(line)
+        if m:
+            result += line
+    return result
+
 def compare_keys(k1, k2):
     "Compare ASCII keys"
-    k1 = k1.split('\n')
-    k2 = k2.split('\n')
-    del k1[1] # remove version lines
-    del k2[1]
-    return k1 != k2
+    # See issue #57: we need to compare only the actual key data,
+    # ignoring things like spurious blank lines
+    return get_key_data(k1) != get_key_data(k2)
 
 class GPGTestCase(unittest.TestCase):
     def setUp(self):
@@ -102,12 +158,12 @@ class GPGTestCase(unittest.TestCase):
         self.gpg = gpg = gnupg.GPG(gnupghome=hd, gpgbinary=GPGBINARY)
         v = gpg.version
         if v:
-            if v >= (2,):
+            if v >= (2,):  # pragma: no cover
                 gpg.options = ['--debug-quick-random']
             else:
                 gpg.options = ['--quick-random']
         self.test_fn = test_fn = 'random_binary_data'
-        if not os.path.exists(test_fn):
+        if not os.path.exists(test_fn):  # pragma: no cover
             data_file = open(test_fn, 'wb')
             data_file.write(os.urandom(5120 * 1024))
             data_file.close()
@@ -237,14 +293,36 @@ class GPGTestCase(unittest.TestCase):
         for _, _, sfp in key_info['subkeys']:
             self.assertTrue(sfp in public_keys.key_map)
             self.assertTrue(public_keys.key_map[sfp] is key_info)
+
+        # now test with sigs=True
+        public_keys_sigs = self.gpg.list_keys(sigs=True)
+        self.assertTrue(is_list_with_len(public_keys_sigs, 1),
+                        "1-element list expected")
+        key_info = public_keys_sigs[0]
+        fp = key_info['fingerprint']
+        self.assertTrue(fp in public_keys_sigs.key_map)
+        self.assertTrue(public_keys_sigs.key_map[fp] is key_info)
+        self.assertTrue(is_list_with_len(key_info['sigs'], 2))
+        for siginfo in key_info['sigs']:
+            self.assertTrue(len(siginfo), 3)
+        for _, _, sfp in key_info['subkeys']:
+            self.assertTrue(sfp in public_keys_sigs.key_map)
+            self.assertTrue(public_keys_sigs.key_map[sfp] is key_info)
+
         private_keys = self.gpg.list_keys(secret=True)
         self.assertTrue(is_list_with_len(private_keys, 1),
                         "1-element list expected")
+        self.assertEqual(len(private_keys.fingerprints), 1)
         # Now do the same test, but using keyring and secret_keyring arguments
+        pkn = 'pubring.gpg'
+        skn = 'secring.gpg'
         hd = os.path.join(os.getcwd(), 'keys')
+        if os.name == 'posix':
+            pkn = os.path.join(hd, pkn)
+            skn = os.path.join(hd, skn)
         gpg = gnupg.GPG(gnupghome=hd, gpgbinary=GPGBINARY,
-                        keyring=os.path.join(hd, 'pubring.gpg'),
-                        secret_keyring=os.path.join(hd, 'secring.gpg'))
+                        keyring=pkn, secret_keyring=skn)
+        logger.debug('Using keyring and secret_keyring arguments')
         public_keys_2 = gpg.list_keys()
         self.assertEqual(public_keys_2, public_keys)
         private_keys_2 = gpg.list_keys(secret=True)
@@ -283,6 +361,20 @@ class GPGTestCase(unittest.TestCase):
                         'Donna Davis <donna.davis@delta.com>'])
         self.assertEqual(actual, expected)
 
+    def test_list_signatures(self):
+        logger.debug("test_list_signatures begins")
+        imported = self.gpg.import_keys(SIGNED_KEYS)
+        keys = self.gpg.list_keys(keys=["18897CA2"])
+        self.assertTrue(is_list_with_len(keys, 1), "importing test signed key")
+        sigs = self.gpg.list_keys(keys=["18897CA2"], sigs=True)[0]['sigs']
+        logger.debug("testing self-signature")
+        self.assertTrue(('BC6F350FB3D04076', 'Joshua Calvert (A test user) <jc@example.com>', '13x') in sigs)
+        logger.debug("testing subkey self-signature")
+        self.assertTrue(('BC6F350FB3D04076', 'Joshua Calvert (A test user) <jc@example.com>', '18x') in sigs)
+        logger.debug("testing other signature")
+        self.assertTrue(('057CCF658C074FDA', 'Winston Smith (A test user) <winston@example.com>', '10x') in sigs)
+        logger.debug("test_list_signatures ends")
+
     def test_scan_keys(self):
         "Test that external key files can be scanned"
         expected = set([
@@ -315,7 +407,7 @@ class GPGTestCase(unittest.TestCase):
         edata = str(gpg.encrypt(data, barbara))
         self.assertNotEqual(data, edata, "Data must have changed")
         ddata = gpg.decrypt(edata, passphrase="bbrown")
-        if data != ddata.data:
+        if data != ddata.data:  # pragma: no cover
             logger.debug("was: %r", data)
             logger.debug("new: %r", ddata.data)
         self.assertEqual(data, ddata.data, "Round-trip must work")
@@ -357,15 +449,18 @@ class GPGTestCase(unittest.TestCase):
                         "Exported key should be public")
         ascii = ascii.replace("\r", "").strip()
         match = compare_keys(ascii, KEYS_TO_IMPORT)
-        if match:
+        if match:  # pragma: no cover
             logger.debug("was: %r", KEYS_TO_IMPORT)
             logger.debug("now: %r", ascii)
         self.assertEqual(0, match, "Keys must match")
         #Generate a key so we can test exporting private keys
         key = self.do_key_generation()
         ascii = gpg.export_keys(key.fingerprint, True)
+        self.assertTrue(isinstance(ascii, gnupg.text_type))
         self.assertTrue(ascii.find("PGP PRIVATE KEY BLOCK") >= 0,
                         "Exported key should be private")
+        binary = gpg.export_keys(key.fingerprint, True, armor=False)
+        self.assertFalse(isinstance(binary, gnupg.text_type))
         logger.debug("test_import_and_export ends")
 
     def test_import_only(self):
@@ -384,7 +479,7 @@ class GPGTestCase(unittest.TestCase):
                         "Exported key should be public")
         ascii = ascii.replace("\r", "").strip()
         match = compare_keys(ascii, KEYS_TO_IMPORT)
-        if match:
+        if match:  # pragma: no cover
             logger.debug("was: %r", KEYS_TO_IMPORT)
             logger.debug("now: %r", ascii)
         self.assertEqual(0, match, "Keys must match")
@@ -406,7 +501,7 @@ class GPGTestCase(unittest.TestCase):
         self.assertTrue(sig, "Good passphrase should succeed")
         self.assertTrue(sig.hash_algo)
         verified = self.gpg.verify(sig.data)
-        if key.fingerprint != verified.fingerprint:
+        if key.fingerprint != verified.fingerprint:  # pragma: no cover
             logger.debug("key: %r", key.fingerprint)
             logger.debug("ver: %r", verified.fingerprint)
         self.assertEqual(key.fingerprint, verified.fingerprint,
@@ -422,10 +517,11 @@ class GPGTestCase(unittest.TestCase):
         try:
             file = gnupg._make_binary_stream(sig.data, self.gpg.encoding)
             verified = self.gpg.verify_file(file)
-        except UnicodeDecodeError: #happens in Python 2.6
+        except UnicodeDecodeError:  # pragma: no cover
+            # sometimes happens in Python 2.6
             from io import BytesIO
             verified = self.gpg.verify_file(BytesIO(sig.data))
-        if key.fingerprint != verified.fingerprint:
+        if key.fingerprint != verified.fingerprint:  # pragma: no cover
             logger.debug("key: %r", key.fingerprint)
             logger.debug("ver: %r", verified.fingerprint)
         self.assertEqual(key.fingerprint, verified.fingerprint,
@@ -439,10 +535,11 @@ class GPGTestCase(unittest.TestCase):
         try:
             file = gnupg._make_binary_stream(sig.data, self.gpg.encoding)
             verified = self.gpg.verify_file(file, self.test_fn)
-        except UnicodeDecodeError: #happens in Python 2.6
+        except UnicodeDecodeError:  # pragma: no cover
+            # sometimes happens in Python 2.6
             from io import BytesIO
             verified = self.gpg.verify_file(BytesIO(sig.data))
-        if key.fingerprint != verified.fingerprint:
+        if key.fingerprint != verified.fingerprint:  # pragma: no cover
             logger.debug("key: %r", key.fingerprint)
             logger.debug("ver: %r", verified.fingerprint)
         self.assertEqual(key.fingerprint, verified.fingerprint,
@@ -458,7 +555,7 @@ class GPGTestCase(unittest.TestCase):
             verified = self.gpg.verify_data(fn, data)
         finally:
             os.unlink(fn)
-        if key.fingerprint != verified.fingerprint:
+        if key.fingerprint != verified.fingerprint:  # pragma: no cover
             logger.debug("key: %r", key.fingerprint)
             logger.debug("ver: %r", verified.fingerprint)
         self.assertEqual(key.fingerprint, verified.fingerprint,
@@ -551,7 +648,7 @@ class GPGTestCase(unittest.TestCase):
             ddata = dfile.read()
             dfile.close()
             data = data.encode(self.gpg.encoding)
-            if ddata != data:
+            if ddata != data:  # pragma: no cover
                 logger.debug("was: %r", data)
                 logger.debug("new: %r", ddata)
             self.assertEqual(data, ddata, "Round-trip must work")
@@ -631,7 +728,7 @@ class GPGTestCase(unittest.TestCase):
         finally:
             shutil.rmtree(workdir)
 
-    def disabled_test_signing_with_uid(self):
+    def disabled_test_signing_with_uid(self):  # pragma: no cover
         "Test that signing with uids works. On hold for now."
         logger.debug("test_signing_with_uid begins")
         key = self.generate_key("Andrew", "Able", "alpha.com")
@@ -653,6 +750,7 @@ TEST_GROUPS = {
                    'test_filenames_with_spaces']),
     'key' : set(['test_deletion', 'test_import_and_export',
                  'test_list_keys_after_generation',
+                 'test_list_signatures',
                  'test_key_generation_with_invalid_key_type',
                  'test_key_generation_with_escapes',
                  'test_key_generation_with_empty_value',
@@ -671,7 +769,7 @@ def suite(args=None):
     if not args or args == ['--no-doctests']:
         result = unittest.TestLoader().loadTestsFromTestCase(GPGTestCase)
         want_doctests = not args
-    else:
+    else:  # pragma: no cover
         tests = set()
         want_doctests = False
         for arg in args:
@@ -688,7 +786,7 @@ def suite(args=None):
 
 def init_logging():
     logging.basicConfig(level=logging.DEBUG, filename="test_gnupg.log",
-                        filemode="w", format="%(asctime)s %(levelname)-5s %(name)-10s %(threadName)-10s %(message)s")
+                        filemode="w", format="%(asctime)s %(levelname)-5s %(name)-10s %(threadName)-10s %(lineno)4d %(message)s")
 
 def main():
     init_logging()
